import { Ok, Err, curry } from '../vendor/kojima/src/index.js'

export class ParseError extends Error {
  constructor(message, state, source) {
    super(message)
    this.state = state
    this.source = source
  }
}

const Tuple = (...values) => Object.freeze(values)
const State = value => Tuple([], Iterator.from(value))

const LowerAlpha = 'abcdefghijklmnopqrstuvwxyz'
const UpperAlpha = LowerAlpha.toUpperCase()
const Alpha = LowerAlpha + UpperAlpha
const Digits = '1234567890'
const Alphanumeric = Alpha + Digits


const tee = (iterator, n = 2) => {
  iterator = Iterator.from(iterator)

  function* gen(current) {
    while (true) {
      if (!current.next) {
        const { done, value } = iterator.next()
        if (done) { return }
        current.next = { value }
      }
      current = current.next
      yield current.value
    }
  }

  return Array(n).fill({}).map(gen)
}

const fork = ([tokens, state]) => {
  const [a, b] = tee(state)
  return Tuple(
    Tuple(tokens.slice(), a),
    Tuple(tokens.slice(), b),
  )
}

export const succeed = (v, [x, y]) => Ok(Tuple(x.concat(v), y))
export const fail = (msg, state, err = undefined) => Err(new ParseError(msg, state, err))
const nth = (n, iter) => iter[n]
const next = state => nth(1, state).next().value
const diff = (a, b) => b.slice(-Math.max(0, b.length - a.length))
const join = curry((delim, val) => val.join(delim))
const mapStr = curry((fn, str) => Array.from(str).map(v => fn(v)))

export const any = (...parsers) => state => {
  for (const parser of parsers) {
    const [original, clone] = fork(state)
    const result = parser(clone)
    if (result.isOk) {
      return result
    }
  }

  return fail('no matching parsers', state)
}

export const anyOf = curry((str, state) => (
  any(...mapStr(char, str))(state)
))

export const seq = (...parsers) => state => {
  let acc = Ok(state)

  for (const parser of parsers) {
    if (acc.isOk) {
      acc = acc.bind(parser)
    } else {
      break
    }
  }

  return acc
}

export const map = curry((fn, parser, state) => {
  return parser(state).bind(result => {
    try {
      const parsed = diff(state[0], result[0])
      const backtrack = Tuple(state[0], result[1])
      return succeed(fn(parsed), backtrack)
    } catch (e) {
      return fail('failed to map', state, e)
    }
  })
})

export const char = curry((ch, state) => (
  next(state) === ch ? succeed(ch, state) : fail(`could not parse ${ch} `, state)
))

export const anyChar = state => {
  const ch = next(state)
  return !!ch ? succeed(ch, state) : fail(`could not parse ${ch}`, state)
}

export const str = curry((str, state) => (
  map(
    join(''),
    seq(...mapStr(char, str))
  )(state)
))

export const digit = anyOf(Digits)
export const lowerAlpha = anyOf(LowerAlpha)
export const upperAlpha = anyOf(UpperAlpha)
export const alpha = anyOf(Alpha)
export const alphanumeric = anyOf(Alphanumeric)

export const maybe = curry((parser, state) => {
  const [original, clone] = fork(state)
  const result = parser(clone)
  return result.isOk ? result : succeed([], original)
})

export const parse = curry((parser, input) => parser(State(input)))

